//
// XMM Science Analysis System
//
// Giuseppe Vacanti (European Space Agency) 
// August 13, 1999
//

#include <Dal.h>
#include <errstr.h>
#include <param.h>
#include <Tracer.h>
#include <ExprColumnGenerator.h>
#include <STime.h>

#include <string>
//#include <limits>
#include <float.h>
#include <sstream>
#include <iomanip>
#include <iostream>

#include "Phaser.h"


static std::string expression(STime tzero, STime refepoch, double refphase, double f, double fdot, double f2dot = 0.)
{
  PROC
  // the expression could possibly be optimized
  stringstream e;
  double tconst = tzero - refepoch;
  e << std::setprecision(15) << "modf ( " << refphase << " + " << f
    << " * ( TIME + " << tconst << " ) + 0.5 * ( TIME + "
    << tconst << " )**2 * " << fdot << " + ( TIME + " << tconst
    << " ) **3 * " << f2dot << " / 6.0)" ;
  errstr << "Expression = " << e.str() << Sas::message(Msg::App, Msg::Verbose);
  RETURN e.str();
}

Phaser::Phaser(STime refepoch, double refphase, double f, double fdot)
  : _refepoch(refepoch),
    _refphase(refphase),
    _f(f),
    _fdot(fdot),
    _f2dot(0.)
{
  PROC
  if(! invariant())
    Sas::error("TooSmall", "The frequency is too small.");
  ENDPROC
}

Phaser::~Phaser() { PROC ENDPROC }

void Phaser::setReferenceEpoch(STime refepoch)
{
  PROC
  _refepoch = refepoch;
  ENDPROC
}

void Phaser::setReferencePhase(double refphase)
{
  PROC
  _refphase = refphase;
  ENDPROC
}

void Phaser::setFrequency(double f)
{
  PROC
   _f = f;
  if(! invariant())
    Sas::error("TooSmall", "The frequency is too small.");
  ENDPROC
}

void Phaser::setFrequencyDot(double fdot)
{
  PROC
  _fdot = fdot;
  ENDPROC
}

STime Phaser::referenceEpoch() const
{
  PROC
  RETURN _refepoch;
}

double Phaser::referencePhase() const
{
  PROC
  RETURN _refphase;
}

double Phaser::frequency() const
{
  PROC
  RETURN _f;
}

double Phaser::frequencyDot() const
{
  PROC
  RETURN _fdot;
}

// Calculated under the assumption that periodTwoDot = 0 (as in Crab)
double Phaser::frequencyTwoDot() const
{
  PROC
  RETURN 2. * _fdot * _fdot / _f;
}    

double Phaser::period() const
{
  PROC
  RETURN 1/_f;
}

double Phaser::periodDot() const
{
  PROC
  RETURN - _fdot/(_f * _f);
}

void Phaser::phase(Table * tab) const
{
  PROC

  map<string, STime::TimeSystem > timesysmap;
  timesysmap["TT"]  = STime::TT;
  timesysmap["UTC"] = STime::UTC;
  timesysmap["TAI"] = STime::TAI;  

  double mjdref = 0;
  if(tab->hasAttribute("MJDREF"))
    mjdref = tab->attribute("MJDREF")->asReal();
  else if(tab->parent()->hasAttribute("MJDREF"))
    mjdref =  tab->parent()->attribute("MJDREF")->asReal();
  double timezero = 0;

  if(tab->hasAttribute("TIMEZERO"))
    timezero = tab->attribute("TIMEZERO")->asReal();
  else if(tab->parent()->hasAttribute("TIMEZERO"))
    timezero =  tab->parent()->attribute("TIMEZERO")->asReal();
  errstr << tab->qualifiedName() << " : MJDREF is " << mjdref
	 << " (" << STime(mjdref) << ")"
	 << Sas::message(Msg::App, Msg::Sparse);
  errstr << tab->qualifiedName() << " : TIMEZERO is " << timezero
	 << Sas::message(Msg::App, Msg::Sparse);

  string timesys = "TT";
  if(!tab->hasAttribute("TIMESYS"))
    errstr << tab->qualifiedName()
	   << " does not specify the time system. TT assumed."
 	   << Sas::warning("NoTimeSystem");
  else
    timesys = tab->attribute("TIMESYS")->asString();

  errstr << tab->qualifiedName() << " : TIMESYS is " << timesys
	 << Sas::message(Msg::App, Msg::Sparse);

  // This is the start of the observation
  STime tzero(mjdref + timezero/STime::secondsPerDay);
  
  // Convert reference epoch to correct time system
  STime::TimeSystem timeSystem = timesysmap[timesys];
  STime convertedRefEpoch;
  
  if (timeSystem != STime::UTC) {
     double mjdTime = referenceEpoch().modifiedJulianDay(timeSystem).full();
	 convertedRefEpoch = STime(mjdTime, STime::MJD);
  }
  else {
	 convertedRefEpoch = referenceEpoch();
  }
  
  errstr << "Input Reference Epoch is " << referenceEpoch().toString(6) 
	 << Sas::message(Msg::App, Msg::Sparse);
  errstr << "Reference Epoch is " << convertedRefEpoch.toString(6) 
	 << Sas::message(Msg::App, Msg::Sparse);
  
  string e = expression(tzero, convertedRefEpoch, referencePhase(),
			frequency(), frequencyDot());

  ExprColumnGenerator cg(e,
			 "PHASE",
			 Dal::Real64);
  cg.doIt(*tab, *tab);

  if(tab->hasAttribute("PHASEXP"))
    tab->deleteAttribute("PHASEXP");
  *tab->addAttribute("PHASEXP") = e;

  ENDPROC
}

void Phaser::print(ostream & os) const
{
  PROC
  os << std::setprecision(15);
  os << "Reference epoch         = " << referenceEpoch() << endl;
  os << "Reference phase         = " << referencePhase() << endl;
  os << "Frequency (Hz)          = " << frequency() << endl;
  os << "Frequency dot (Hz/s)    = " << frequencyDot() << endl;
  os << "Frequency 2dot (Hz/s^2) = " << frequencyTwoDot() << endl;
  os << "Period (s)              = " << period() << endl;
  os << "Period dot (s/s)        = " << periodDot() << endl;
  ENDPROC;
}

bool Phaser::invariant() const
{
  PROC
  RETURN (std::fabs(_f) > LDBL_EPSILON);
}
